// kinetracker.cpp
//   Developer by Eiichiro Ito(ghc02331@nifty.com)
//
// Based on openni_tracker originally developed by Tim Field
//   http://ros.org/wiki/openni_tracker

#include <ros/ros.h>
#include <ros/package.h>

#include <XnOpenNI.h>
#include <XnCodecIDs.h>
#include <XnCppWrapper.h>

#include "kinetracker/Skelton.h"

using std::string;

xn::Context        g_Context;
xn::DepthGenerator g_DepthGenerator;
xn::UserGenerator  g_UserGenerator;

XnBool g_bNeedPose   = FALSE;
XnChar g_strPose[20] = "";

void XN_CALLBACK_TYPE
User_NewUser(xn::UserGenerator& generator, XnUserID nId, void* pCookie)
{
  ROS_INFO("New User %d", nId);

  if (g_bNeedPose)
    g_UserGenerator.GetPoseDetectionCap().StartPoseDetection(g_strPose, nId);
  else
    g_UserGenerator.GetSkeletonCap().RequestCalibration(nId, TRUE);
}

void XN_CALLBACK_TYPE
User_LostUser(xn::UserGenerator& generator, XnUserID nId, void* pCookie)
{
  ROS_INFO("Lost user %d", nId);
}

void XN_CALLBACK_TYPE
UserCalibration_CalibrationStart(xn::SkeletonCapability& capability, XnUserID nId, void* pCookie)
{
  ROS_INFO("Calibration started for user %d", nId);
}

void XN_CALLBACK_TYPE
UserCalibration_CalibrationEnd(xn::SkeletonCapability& capability, XnUserID nId, XnBool bSuccess, void* pCookie)
{
  if (bSuccess) {
    ROS_INFO("Calibration complete, start tracking user %d", nId);
    g_UserGenerator.GetSkeletonCap().StartTracking(nId);
  }
  else {
    ROS_INFO("Calibration failed for user %d", nId);
    if (g_bNeedPose)
      g_UserGenerator.GetPoseDetectionCap().StartPoseDetection(g_strPose, nId);
    else
      g_UserGenerator.GetSkeletonCap().RequestCalibration(nId, TRUE);
  }
}

void XN_CALLBACK_TYPE
UserPose_PoseDetected(xn::PoseDetectionCapability& capability, XnChar const* strPose, XnUserID nId, void* pCookie)
{
  ROS_INFO("Pose %s detected for user %d", strPose, nId);
  g_UserGenerator.GetPoseDetectionCap().StopPoseDetection(nId);
  g_UserGenerator.GetSkeletonCap().RequestCalibration(nId, TRUE);
}

void
setJoint(XnUserID const& user, XnSkeletonJoint const& joint, geometry_msgs::Vector3& pos)
{
    XnSkeletonJointPosition joint_position;

    g_UserGenerator.GetSkeletonCap().GetSkeletonJointPosition(user, joint, joint_position);
    pos.x = -joint_position.position.X / 1000.0;
    pos.y = joint_position.position.Y / 1000.0;
    pos.z = joint_position.position.Z / 1000.0;
}

void
publishTransforms(const ros::Publisher& publisher)
{
  XnUserID users[15];
  XnUInt16 users_count = 15;
  g_UserGenerator.GetUsers(users, users_count);

  for (int i = 0; i < users_count; ++i) {
    XnUserID user = users[i];
    if (!g_UserGenerator.GetSkeletonCap().IsTracking(user))
      continue;

    kinetracker::Skelton msg;
    msg.id = user;
    setJoint(user, XN_SKEL_HEAD, msg.head);
    setJoint(user, XN_SKEL_NECK, msg.neck);
    setJoint(user, XN_SKEL_TORSO, msg.torso);
    setJoint(user, XN_SKEL_LEFT_SHOULDER, msg.left_shoulder);
    setJoint(user, XN_SKEL_LEFT_ELBOW, msg.left_elbow);
    setJoint(user, XN_SKEL_LEFT_HAND, msg.left_hand);
    setJoint(user, XN_SKEL_RIGHT_SHOULDER, msg.right_shoulder);
    setJoint(user, XN_SKEL_RIGHT_ELBOW, msg.right_elbow);
    setJoint(user, XN_SKEL_RIGHT_HAND, msg.right_hand);
    setJoint(user, XN_SKEL_LEFT_HIP, msg.left_hip);
    setJoint(user, XN_SKEL_LEFT_KNEE, msg.left_knee);
    setJoint(user, XN_SKEL_LEFT_FOOT, msg.left_foot);
    setJoint(user, XN_SKEL_RIGHT_HIP, msg.right_hip);
    setJoint(user, XN_SKEL_RIGHT_KNEE, msg.right_knee);
    setJoint(user, XN_SKEL_RIGHT_FOOT, msg.right_foot);
    publisher.publish(msg);
  }
}

#define CHECK_RC(nRetVal, what)						\
  if (nRetVal != XN_STATUS_OK)						\
    {									\
      ROS_ERROR("%s failed: %s", what, xnGetStatusString(nRetVal));	\
      return nRetVal;							\
    }

int main(int argc, char **argv)
{
  ros::init(argc, argv, "kinetracker");
  ros::NodeHandle nh;

  ROS_INFO("Start kinetracker");

  string configFilename = ros::package::getPath("kinetracker") + "/kinetracker.xml";
  XnStatus nRetVal = g_Context.InitFromXmlFile(configFilename.c_str());
  CHECK_RC(nRetVal, "InitFromXml");

  nRetVal = g_Context.FindExistingNode(XN_NODE_TYPE_DEPTH, g_DepthGenerator);
  CHECK_RC(nRetVal, "Find depth generator");

  nRetVal = g_Context.FindExistingNode(XN_NODE_TYPE_USER, g_UserGenerator);
  if (nRetVal != XN_STATUS_OK) {
    nRetVal = g_UserGenerator.Create(g_Context);
    CHECK_RC(nRetVal, "Find user generator");
  }

  ROS_INFO("[d]");

  if (!g_UserGenerator.IsCapabilitySupported(XN_CAPABILITY_SKELETON)) {
    ROS_INFO("Supplied user generator doesn't support skeleton");
    return 1;
  }

  ROS_INFO("[e]");

  XnCallbackHandle hUserCallbacks;
  g_UserGenerator.RegisterUserCallbacks(User_NewUser, User_LostUser, NULL, hUserCallbacks);

  ROS_INFO("[f]");

  XnCallbackHandle hCalibrationCallbacks;
  g_UserGenerator.GetSkeletonCap().RegisterCalibrationCallbacks(UserCalibration_CalibrationStart, UserCalibration_CalibrationEnd, NULL, hCalibrationCallbacks);

  ROS_INFO("[g]");

  if (g_UserGenerator.GetSkeletonCap().NeedPoseForCalibration()) {
    g_bNeedPose = TRUE;
    if (!g_UserGenerator.IsCapabilitySupported(XN_CAPABILITY_POSE_DETECTION)) {
      ROS_INFO("Pose required, but not supported");
      return 1;
    }

    XnCallbackHandle hPoseCallbacks;
    g_UserGenerator.GetPoseDetectionCap().RegisterToPoseCallbacks(UserPose_PoseDetected, NULL, NULL, hPoseCallbacks);

    g_UserGenerator.GetSkeletonCap().GetCalibrationPose(g_strPose);
  }

  ROS_INFO("[h]");

  g_UserGenerator.GetSkeletonCap().SetSkeletonProfile(XN_SKEL_PROFILE_ALL);

  ROS_INFO("[k]");

  nRetVal = g_Context.StartGeneratingAll();
  CHECK_RC(nRetVal, "StartGenerating");

  ROS_INFO("[l]");

  ros::Rate r(30);

  ROS_INFO("[m]");

  ros::NodeHandle n;
  ros::Publisher tracker_pub = n.advertise<kinetracker::Skelton>("/skelton", 1000);

  while (ros::ok()) {
    g_Context.WaitAndUpdateAll();
    publishTransforms(tracker_pub);
    r.sleep();
  }

  ROS_INFO("[Shutdown kinetracker]");

  g_Context.Shutdown();
  return 0;
}
